pragma solidity >=0.5.11;

contract Casino {
    //Represents a single bet
    struct Bet {
        int prediction; // The prediction for the winner
        int wager; // The amount of money the Bettor put down on this BetPrediction
    }

    Game currentGame; //The game that is currently being played
    int casinoPot; // Number of tokens owned by the Casino

    // Assume there's code that lets people buy and sell tokens. You don't need to write it.
    mapping (address => int) tokenBalances;
    mapping (address => Bet) bets;
    mapping (address => bool) betsTaken;

    constructor(Game g) public {
        currentGame = g;
        casinoPot = 1000; // Assume the casino starts with plenty of tokens
    }


    // Let Bettors collect winnings. We will return the tokens (if applicable) as an int value.
    function collectWinnings() public {
        require(currentGame.state() == Game.State.FinishedPlaying, "Wrong initial state.");

        Bet memory b = bets[msg.sender];
        int win = 0;
        
        if (currentGame.predictionMatchesOutcome(b.prediction)) {
            win = 2*b.wager;
            casinoPot = casinoPot-win;
            require(casinoPot >= 0, "Casino ran out of money!");
        }
        
        // either way, clear this person's bet and award the win value
        betsTaken[msg.sender] = false;
        delete bets[msg.sender];
        tokenBalances[msg.sender] += win;
    }

    // Function to let Bettors place bets
    function placeBet(int w, int p) public {
        if (currentGame.state() != Game.State.BeforePlay) {
            return;
        }

        if (betsTaken[msg.sender]) {
            Bet memory b = bets[msg.sender];
            tokenBalances[msg.sender] += b.wager;
        }

        if (w > tokenBalances[msg.sender]) {
            return;
        }

        // otherwise good to go!
        bets[msg.sender] = Bet(p, w);
        betsTaken[msg.sender] = true;
        tokenBalances[msg.sender] -= w;
    }
}

//Represents a generic game
contract Game {
    enum State { BeforePlay, Playing, FinishedPlaying}
    State public state;
    int outcome; // In FinishedPlaying state, stores information about the outcome of the game.

    constructor() public {
        state = State.BeforePlay;
    }

    //Start the game
    function startPlaying() public {
        require(state == State.BeforePlay, "Wrong initial state.");
        state = State.Playing;
    }

    //Finish the game
    function finishPlaying() public {
        require(state == State.Playing, "Wrong initial state.");
        outcome = 42; // Just assume something simple for now.
        state = State.FinishedPlaying;
    }

    //Returns the outcome of the game
    function predictionMatchesOutcome(int prediction) public view returns (bool) {
        require(state == State.FinishedPlaying, "Wrong initial state.");
        //...
        return prediction == outcome;
    }
}