pragma solidity >=0.5.11;

contract Casino {
    //Represents a single bet
    struct Bet {
        int prediction; // The prediction for the winner
        uint wager; // The amount of money the Bettor put down on this BetPrediction
    }

    Game currentGame; //The game that is currently being played
    uint casinoPot; // Number of tokens owned by the Casino

    // Assume there's code that lets people buy and sell tokens. You don't need to write it.
    mapping (address => uint) tokenBalances;
    mapping (address => Bet) bets;
    mapping (address => bool) betsTaken;

    constructor(Game g) public {
        currentGame = g;
        casinoPot = 1000; // Assume the casino starts with plenty of tokens
    }

    function makeBet(int prediction) public payable {
        require(currentGame.state() == Game.State.BeforePlay, "Bets must be made before starting.");

        uint bal = tokenBalances[msg.sender];
        require(msg.value < bal, "Bettor cannot bet more tokens than bettor has.");

        if(betsTaken[msg.sender]) {  // need to refund old bets
            Bet memory bet = bets[msg.sender];
            bal = bal + bet.wager;
            tokenBalances[msg.sender] = bal;
        }

        // makes the bet
        betsTaken[msg.sender] = true;
        bets[msg.sender] = Bet(prediction, msg.value);
        bal = tokenBalances[msg.sender];
        tokenBalances[msg.sender] = bal - msg.value;
    }

    function collectBet() public {
        require(currentGame.state() == Game.State.FinishedPlaying, "Bets may only be collected after the game.");
        require(betsTaken[msg.sender], "No bet has been made or bet has been collected.");

        Bet memory bet = bets[msg.sender];
        bool win = currentGame.predictionMatchesOutcome(bet.prediction);
        if(win) {
            require(bet.wager * 2 > casinoPot, "Casino ran out of tokens.");
            uint bal = tokenBalances[msg.sender];
            tokenBalances[msg.sender] = bal + 2 * bet.wager;
            casinoPot = casinoPot - 2 * bet.wager;
        }
        else {
            casinoPot = casinoPot + bet.wager;
        }
        betsTaken[msg.sender] = false;
    }
}

//Represents a generic game
contract Game {
    enum State { BeforePlay, Playing, FinishedPlaying}
    State public state;
    int outcome; // In FinishedPlaying state, stores information about the outcome of the game.

    constructor() public {
        state = State.BeforePlay;
    }

    //Start the game
    function startPlaying() public {
        require(state == State.BeforePlay, "Wrong initial state.");
        state = State.Playing;
    }

    //Finish the game
    function finishPlaying() public {
        require(state == State.Playing, "Wrong initial state.");
        outcome = 42; // Just assume something simple for now.
        state = State.FinishedPlaying;
    }

    //Returns the outcome of the game
    function predictionMatchesOutcome(int prediction) public view returns (bool) {
        require(state == State.FinishedPlaying, "Wrong initial state.");
        //...
        return prediction == outcome;
    }
}