pragma solidity >=0.5.11;

contract Casino {
    //Represents a single bet
    struct Bet {
        int prediction; // The prediction for the winner
        uint wager; // The amount of money the Bettor put down on this BetPrediction
    }

    Game currentGame; //The game that is currently being played
    uint casinoPot; // Number of tokens owned by the Casino
    bool isPlaying; // not sure how to access the state of the currentGame otherwise (or more accurately, how to compare it to elements of its state enum)

    // Assume there's code that lets people buy and sell tokens. You don't need to write it.
    mapping (address => uint) tokenBalances;
    mapping (address => Bet) bets;
    mapping (address => bool) betsTaken;

    constructor(Game g) public {
        currentGame = g;
        casinoPot = 1000; // Assume the casino starts with plenty of tokens
    }

    function startGame() public {
        isPlaying = true;
        currentGame.startPlaying();
    }

    function endGame() public {
        isPlaying = false;
        currentGame.finishPlaying();
    }

    function collectWinnings() public {
        assert(!isPlaying);
        Bet storage b = bets[msg.sender];
        if(currentGame.predictionMatchesOutcome(b.prediction)) {
            if(2 * b.wager >= casinoPot) {
                tokenBalances[msg.sender] += 2 * b.wager;
                casinoPot -= 2 * b.wager;
            } else {
                revert("Casino cannot pay");
            }
        } else {
            casinoPot += b.wager; // otherwise take their tokens
        }

        // could return the tokens as money here with transfer()?
    }

    // assuming that there are already token balances for each player
    // (they bought them earlier, no need to make the function payable)
    function makeBet(int pred, uint w) public {
        require(!isPlaying, "Bets can only be made before the game starts");
        // can have one active bet per game,
        // otherwise reimburse old bet and replace it
        if(betsTaken[msg.sender]) {
            Bet storage oldBet = bets[msg.sender];
            tokenBalances[msg.sender] += oldBet.wager;
        }
        // cannot wager more tokens than they own
        assert(tokenBalances[msg.sender] >= w);
        Bet storage b = bets[msg.sender];
        b.prediction = pred;
        b.wager = w;
        betsTaken[msg.sender] = true;
        // remove from their balance
        tokenBalances[msg.sender] -= w;
    }
    //TODO: Add your code here.
}

//Represents a generic game
contract Game {
    enum State { BeforePlay, Playing, FinishedPlaying}
    State public state;
    int outcome; // In FinishedPlaying state, stores information about the outcome of the game.

    constructor() public {
        state = State.BeforePlay;
    }

    //Start the game
    function startPlaying() public {
        require(state == State.BeforePlay, "Wrong initial state.");
        state = State.Playing;
    }

    //Finish the game
    function finishPlaying() public {
        require(state == State.Playing, "Wrong initial state.");
        outcome = 42; // Just assume something simple for now.
        state = State.FinishedPlaying;
    }

    //Returns the outcome of the game
    function predictionMatchesOutcome(int prediction) public view returns (bool) {
        require(state == State.FinishedPlaying, "Wrong initial state.");
        //...
        return prediction == outcome;
    }
}