pragma solidity >=0.5.11;

contract Casino {
    //Represents a single bet
    struct Bet {
        int prediction; // The prediction for the winner
        uint wager; // The amount of money the Bettor put down on this BetPrediction
    }

    Game currentGame; //The game that is currently being played
    uint casinoPot; // Number of tokens owned by the Casino

    // Assume there's code that lets people buy and sell tokens. You don't need to write it.
    mapping (address => uint) tokenBalances;
    mapping (address => Bet) bets;
    mapping (address => bool) betsTaken;

    constructor(Game g) public {
        currentGame = g;
        casinoPot = 1000; // Assume the casino starts with plenty of tokens
    }

    function makeBet(int prediction) public payable {
        require(currentGame.state() == Game.State.BeforePlay, "Wrong initial state.");

        if (betsTaken[msg.sender] == true) {
            require(tokenBalances[msg.sender] + bets[msg.sender].wager > msg.value, "Bettor does not have enough tokens.");
            msg.sender.transfer(bets[msg.sender].wager);
            casinoPot -= bets[msg.sender].wager;
        } else {
            require(tokenBalances[msg.sender] > msg.value, "Bettor does not have enough tokens.");
        }

        Bet memory b;
        b.prediction = prediction;
        b.wager = msg.value;

        bets[msg.sender] = b;
        betsTaken[msg.sender] = true;

        tokenBalances[msg.sender] -= b.wager;
        casinoPot += msg.value;
    }

    function collectWinnings() public payable {
        require(currentGame.state() == Game.State.FinishedPlaying, "Wrong initial state.");
        require(betsTaken[msg.sender] == true, "No bets taken by msg sender.");

        Bet memory b = bets[msg.sender];
        if (currentGame.predictionMatchesOutcome(b.prediction)) {
            require(b.wager*2 < casinoPot, "Casino can't pay!");
            
            msg.sender.transfer(b.wager*2);
            casinoPot -= b.wager*2;
            tokenBalances[msg.sender] += b.wager*2;
        }
    }
}

//Represents a generic game
contract Game {
    enum State { BeforePlay, Playing, FinishedPlaying}
    State public state;
    int outcome; // In FinishedPlaying state, stores information about the outcome of the game.

    constructor() public {
        state = State.BeforePlay;
    }

    //Start the game
    function startPlaying() public {
        require(state == State.BeforePlay, "Wrong initial state.");
        state = State.Playing;
    }

    //Finish the game
    function finishPlaying() public {
        require(state == State.Playing, "Wrong initial state.");
        outcome = 42; // Just assume something simple for now.
        state = State.FinishedPlaying;
    }

    //Returns the outcome of the game
    function predictionMatchesOutcome(int prediction) public view returns (bool) {
        require(state == State.FinishedPlaying, "Wrong initial state.");
        //...
        return prediction == outcome;
    }
}